/*
    Ishtar - A network transparent read/write interface with
    probing and enumeration support.
    settings part : settings file parsing and writing
    Copyright (C) 2003-2005 Stephane Magnenat <nct@ysagoon.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
	Settings
	
	Provides setting registrar that is itself an Ishtar server
*/

#include <vector>
#include <fstream>
#include <sstream>
#include <stack>
#include "settings.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

/*!	\file settings.cpp
	\brief Implementation of Ishtar settings classes.
*/

namespace Ishtar
{
	//! A parser that is able to create a table of settings from a file or a string.
	struct Parser
	{
		//! Tokens for lexer
		struct Token
		{
			//! Type of token
			enum Type
			{
				END_OF_INPUT = 0, //!< no more data in input stream
				OPEN_BRACKET, //!< opening bracker ({)
				CLOSE_BRACKET, //!< closing bracker ({)
				EQUAL, //!< equal signe (=)
				SEMI_COLON, //!< semi-colon (;)
				VAL, //!< a string stored in val representing a value
				NONE //!< no type
			} type; //!< type of this token
			
			std::string val; //!< Value of the token if any
			
			//! Create a token with a given type and an optional initial value
			Token(Type type, const std::string &val = "")
			{
				this->type = type;
				this->val = val;
			}
		};
		
		std::ifstream::int_type next; //!< the most recently read character
		std::ifstream ifs; //!< active stream when using file
		std::istringstream iss; //!< active stream when using string
		std::istream *is; //!< stream interface pointing to the correct stream
		Token token; //!< the most recently read token
		size_t line; //!< line of reading in the source file
		size_t column; //!< column of reading in the source file
		
		//! Parse content of file filename
		Parser(const std::string &fileName) :
			ifs(fileName.c_str()),
			token(Token::NONE)
		{
			line = 1;
			column = 0;

			is = &ifs;
			
			if (ifs)
				nextChar();
		}

		//! Parse content of string config
		Parser(const std::string &config, bool isAString) :
			iss(config),
			token(Token::NONE)
		{
			line = 1;
			column = 0;

			is = &iss;

			if (iss)
				nextChar();
		}

		//! Return next character, increment line and column position
		std::ifstream::int_type nextChar(void)
		{
			next = is->get();
			if (next == '\n')
			{
				column = 0;
				line++;
			}
			else
				column++;
			return next;
		}
		
		//! Return the next token, return Token::END_OF_INPUT if no more token
		Token nextToken(void)
		{
			if (!is->eof())
			{
				switch (next)
				{
					case '/':
					nextChar();
					if (next == '/')
					{
						// single line comment
						while ((next != '\n') && (next != '\r') && !is->eof())
							nextChar();
						while (((next == '\n') || (next == '\r')) && !is->eof())
							nextChar();
					}
					else if (next == '*')
					{
						// multi line comment
						nextChar();
						while ((next != '*' || (nextChar(), next != '/')) && !is->eof())
							nextChar();
						nextChar();
						
						/* More human readable version.
						nextChar();
						while (!is->eof())
						{
							while ((next != '*') && !!is->eof())
								nextChar();
							nextChar();
							if (next == '/')
							{
								nextChar();
								break;
							}
						}*/
					}
					else
					{
						std::cerr << "!!! Scan (lexer) error @ " << line << ':' << column << " : / or * expected" << std::endl;
					}
					token = nextToken();
					break;
					
					case '{':
					nextChar();
					token = Token(Token::OPEN_BRACKET);
					break;
					
					case '}':
					nextChar();
					token = Token(Token::CLOSE_BRACKET);
					break;
					
					case '=':
					nextChar();
					token = Token(Token::EQUAL);
					break;
					
					case ';':
					nextChar();
					token = Token(Token::SEMI_COLON);
					break;
					
					case '"':
					{
						std::string tempValue;
						while ((nextChar() != '"') && !is->eof())
						{
							tempValue += static_cast<std::string::value_type>(next);
						}
						nextChar();
						token = Token(Token::VAL, tempValue);
					}
					break;
					
					case ' ':
					case '\t':
					case '\n':
					case '\r':
					while (std::isspace(next))
						nextChar();
					token = nextToken();
					break;
					
					default:
					{
						if (isalnum(next) || (next == '.') || (next == '-'))
						{
							std::string tempValue;
							do
							{
								tempValue += static_cast<std::string::value_type>(next);
								nextChar();
							}
							while (isalnum(next) || (next == '.'));
							token = Token(Token::VAL, tempValue);
						}
						else
						{
							std::cerr << "!!! Scan (lexer) error @ " << line << ':' << column << " : character or number or . expected" << std::endl;
							token = nextToken();
						}
					}
					break;
				}
			}
			else
			{
				token = Token(Token::END_OF_INPUT);
			}
			return token;
		}
		
		//! Macro that return false and an error message if parser encountered EOF
		#define CHECK_NOT_EOF \
		if (token.type == Token::END_OF_INPUT) \
		{ \
			 std::cerr << "!!! Scan (parser) error @ " << line << ':' << column << " : unexpected EOF" << std::endl; \
			 return false; \
		}
		
		//! Macro that return false and an error message if anything else then a value is encountered
		#define CHECK_VAL \
		if (token.type != Token::VAL) \
		{ \
			std::cerr << "!!! Scan (parser) error @ " << line << ':' << column << " : value expected" << std::endl; \
			return false; \
		}
		
		//! Parse and fill table with results
		bool parse(ParserTable *table)
		{
			if (!*is)
				return false;
			
			std::vector<std::string> levels;
			std::string id;
			std::string fullId;
			std::stack<unsigned> autovectors; // stack of unsigned used for implicit vectors. If value is 0, the level is not considered using autovector
			
			assert(table);
			
			nextToken();
			while (token.type != Token::END_OF_INPUT)
			{
				CHECK_NOT_EOF
				if (token.type == Token::CLOSE_BRACKET)
				{
					if (levels.size())
					{
						// when using autovector, add a count member
						if (autovectors.top() > 0)
						{
							std::ostringstream oss;
							oss << autovectors.top();
							ParsedValues pv;
							pv.push_back(oss.str());
							(*table)[fullId + ".count"] = pv;
						}
						
						autovectors.pop();
						levels.pop_back();
						fullId.clear();
						
						std::vector<std::string>::size_type size = levels.size();
						std::vector<std::string>::size_type j = 0;
						if (size != 0)
						{
							for (std::vector<std::string>::const_iterator i = levels.begin(); i != levels.end(); ++i)
							{
								fullId += *i;
								if (++j < size)
									fullId += ".";
							}
						}
					}
					else
					{
						std::cerr << "!!! Scan (parser) error @ " << line << ':' << column << " : bracket closed but none was opened" << std::endl;
					}
					nextToken();
				}
				else
				{
					if (token.type == Token::OPEN_BRACKET)
					{
						// autovector entry
						std::ostringstream oss;
						oss << autovectors.top()++;
						id = oss.str();
					}
					else
					{
						// normal entry, have an id
						CHECK_VAL
						id = token.val;
						
						nextToken();
						CHECK_NOT_EOF
					}
					
					if (token.type == Token::EQUAL)
					{
						nextToken();
						CHECK_NOT_EOF
						CHECK_VAL
						
						ParsedValues pv;
						
						while (token.type != Token::SEMI_COLON)
						{
							CHECK_VAL
							pv.push_back(token.val);
							nextToken();
						}
						std::string path = fullId;
						if (fullId.length())
							path += ".";
						path += id;
						(*table)[path] = pv;
						nextToken();
					}
					else if (token.type == Token::OPEN_BRACKET)
					{
						if (fullId.length())
							fullId += ".";
						fullId += id;
						levels.push_back(id);
						autovectors.push(0);
						nextToken();
						CHECK_NOT_EOF
					}
					else
					{
						std::cerr << "!!! Scan (parser) error @ " << line << ':' << column << " : = or { expected" << std::endl;
						return false; 
					}
				}
			}
			
			if (levels.size())
			{
				std::cerr << "!!! Scan (parser) error @ " << line << ':' << column << "End of file called while all } are not closed" << std::endl;
				return false;
			}
			
			return true;
		}
	};
	
	Settings::Settings(UInt16 port) :
		ServicesServer(port)
	{
	}
	
	Settings::~Settings()
	{
	}
	
	void Settings::loadConfigFile(const std::string &name)
	{
		struct stat stat_info;
		if (stat(name.c_str(), &stat_info) == 0)
		{
			std::cout << "Loading config file " << name << " ... " << std::endl;
		
			Parser p(name);
			if (!p.parse(&parsedTable))
			{
				std::cerr << "!!! Can't read config file or file contains errors" << std::endl;
			}
		}
		else
			std::cerr << "Config file " << name << " does not exists" << std::endl;
	}
	
	void Settings::loadConfigString(const std::string &config)
	{
		Parser p(config, true);
		if(!p.parse(&parsedTable))
		{
			std::cerr << "!!! Can't read config string or string contains errors" << std::endl;
		}
	}

	void Settings::dumpTable(void)
	{
		// dump
		std::cout << "Content of settings table :\n";
		for (ParserTable::const_iterator i=parsedTable.begin(); i!=parsedTable.end(); ++i)
		{
			std::cout << i->first << " : ";
			for (ParsedValues::const_iterator j=i->second.begin(); j!=i->second.end(); ++j)
			{
				std::cout << *j << " ";
			}
			std::cout << "\n";
		}
		std::cout << std::endl;
	}
	
	void Settings::saveConfigFile(const std::string &name)
	{
		std::ofstream ofs(name.c_str());
		if (ofs.good())
		{
			std::cout << "Saving config file " << name << " ... " << std::endl;
			for (ParserTable::const_iterator i=parsedTable.begin(); i!=parsedTable.end(); ++i)
			{
				ofs << i->first << " = ";
				for (ParsedValues::const_iterator j=i->second.begin(); j!=i->second.end(); ++j)
				{
					const std::string &s = *j;
					
					// look if we need quotes
					bool needQuotes = false;
					for (size_t k=0; k<s.size(); k++)
					{
						if (!isalnum(s[k]) && !(s[k] == '.') && !(s[k] == '-'))
						{
							needQuotes = true;
							break;
						}
					}
					
					// dump
					if (needQuotes)
						ofs << '\"' << *j << "\" ";
					else
						ofs << *j << " ";
				}
				ofs << ";\n";
			}
		}
		else
			std::cerr << "Can't open config file " << name << " for writing" << std::endl;
	}
	
	//! instance of settings
	Settings *settings = NULL;
	
	void initSettings(int argc, char *argv[])
	{
		UInt16 port = 1234;
		for (int  i=1; i<argc; i++)
		{
			if ((strcmp(argv[i], "-p") == 0) && (i+1 < argc))
				port = atoi(argv[i+1]);
		}
		settings = new Settings(port);
	}
	
	void cleanupSettings(void)
	{
		delete settings;
	}
}
