/*
    Ishtar - A network transparent read/write interface with
    probing and enumeration support.
    settings part : settings file parsing and writing
    Copyright (C) 2003-2005 Stephane Magnenat <nct@ysagoon.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
	Settings
	
	Provides setting registrar that is itself an Ishtar server
*/

#ifndef __ISHTAR_SETTINGS_H
#define __ISHTAR_SETTINGS_H

#include <ishtar/network.h>
#include <ishtar/services.h>
#include <sstream>
#include <set>
#include <string>
#include <valarray>
#include <iterator>

/*!	\file settings.h
	\brief Interface of Ishtar settings classes.
*/

/*!	\mainpage Ishtar settings library reference documentation

	\section intro Introduction
	Ishtar is a general network support and an interprocess setting access library.
	Ishtar is composed of three differents modules.
	This modules is libishtarsettings.
	
	Built upon libishtarservices, it provides a special type of C++ variables:
	\li whose initial values are read from a config file.
	\li that are accessible through the network using Ishtar services protocol (see libishtarservices)
	
	Those variables can thus be modified using any Ishtar client.
	
	\section installation Installation
	The installation procedure is explained step by step in the \ref InstallGuide.
	
	\section usage Usage
	A quick start to most important classes is given in the \ref Cookbook.
	
	\section feedback Feedback
	If you have any comments or suggestions, do not hesitate to send them to
	Stephane Magnenat <nct@ysagoon.com>
	
	\section license License
	This program is free software released under the GNU General Public License version 2.
*/

/*! \class Service
	\brief A service that can be accessed through the network. See libishtarservices documentation for more details.
*/


//! Ishtar is the namespace of the Ishtar class library
namespace Ishtar
{
	//! Value, once parsed, are a vector for strings
	typedef std::vector<std::string> ParsedValues;
	//! The table containing the result of parsing. A map of key / values
	typedef std::map<std::string, ParsedValues> ParserTable;
	
	//! The Settings is a ServiceServer which has support for config files
	class Settings: public ServicesServer
	{
	public:
		//! The settings in string once parsed
		ParserTable parsedTable;
	
	public:
		//! Constructor, serve on port
		Settings(UInt16 port);
		//! Destructor
		virtual ~Settings();
		
		//! Load and parse config file name
		void loadConfigFile(const std::string &name);
		//! Parse a config string that does not come from a file
		void loadConfigString(const std::string &config);
		//! Print the parsing result
		void dumpTable(void);
		//! Save config in file name
		void saveConfigFile(const std::string &name);
	};
	
	
	//! Unique global instance of Settings
	extern Settings *settings;
	
	//! Create the settings object using arguments from argv[]
	void initSettings(int argc = 0, char *argv[] = NULL);
	
	//! Delete the settings object
	void cleanupSettings(void);
	
	//! An Ishtar Service that is pointed by one or more Variable and expose its content through the network.
	template<typename T, bool ro, Size s>
	class GenericVarService:public Service
	{
	private:
		//! The number of Variable that point to this service.
		unsigned refCount;
		
	public:
		//! An array of size s of values of type T
		T v[s];
		
	public:
		//! Constructor, build a service for name
		GenericVarService(const std::string &name)
		{
			description.name = name;
			description.type = TypeResolver<T>::type;
			description.flags = ro ? ServiceDescription::READ_ONLY : 0;
			description.length = s;
			refCount = 1;
		}
		
		//! Read values from the service.
		virtual ValueVector *getValues(void)
		{
			ValueVectorTemplate<T> *vvt = new ValueVectorTemplate<T>(s);
			std::copy(&v[0], &v[s], vvt->value.begin());
			return vvt;
		}
		
		//! Write values to the service.
		virtual void setValues(ValueVector *values)
		{
			ValueVectorTemplate<T> *vvt = dynamic_cast<ValueVectorTemplate<T> *>(values);
			assert(vvt->value.size() == s);
			if (vvt)
			{
				std::copy(vvt->value.begin(), vvt->value.end(), &v[0]);
			}
			else
			{
				const std::string &name = description.name;
				std::cerr << "!!! Wrong type passed to " << name << "::setValues" << std::endl;
			}
		}
		
		//! Decrement the number of Variable pointing to this service. If zero, garbage collect itself
		void decCount() { refCount--; if (refCount==0) settings->eraseService(this); }
		//! Increment the number of Variable pointing to this service
		void incCount() { refCount++; }
	};
	
	//! An Ishtar Service that contains a string and is pointed by one or more Variable and expose its string through the network as a CharValueVector.
	template<bool ro, Size s>
	class GenericVarService<std::string, ro, s>:public Service
	{
	private:
		//! The number of Variable that point to this service.
		unsigned refCount;
		
	public:
		//! The string value
		std::string v;
		
	public:
		//! Constructor, build a service for name
		GenericVarService(const std::string &name)
		{
			description.name = name;
			description.type = ValueVector::UCHAR;
			description.flags = ro ? ServiceDescription::READ_ONLY : 0;
			description.length = 0;
			refCount = 1;
		}
		
		//! Read values from the service.
		virtual ValueVector *getValues(void)
		{
			return new CharValueVector(v);
		}
		
		//! Write values to the service.
		virtual void setValues(ValueVector *values)
		{
			CharValueVector *vvt = dynamic_cast<CharValueVector *>(values);
			if (vvt)
			{
				v = *vvt;
			}
			else
			{
				const std::string &name = description.name;
				std::cerr << "!!! Wrong type passed to " << name << "::setValues" << std::endl;
			}
		}
		
		//! Decrement the number of Variable pointing to this service. If zero, garbage collect itself
		void decCount() { refCount--; if (refCount==0) settings->eraseService(this); }
		//! Increment the number of Variable pointing to this service
		void incCount() { refCount++; }
	};
	
	//! A variable for a scalar value that behaves like a standard C++ variable excepted that its content is stored in a GenericVarService pointed by ptr and is accessible through network
	template<typename T, bool ro = false, Size s = 1, bool warningWhenInitializedWithDefaultValue = true>
	class Variable
	{
	private:
		//! The service really containing the value
		GenericVarService<T, ro, s> *ptr;
		
	public:
		//! Constructor, associate this variable with name
		Variable(const std::string &name, T defaultValue = T())
		{
			associate(name, defaultValue);
		}
		
		//! Destructor
		~Variable()
		{
			assert(ptr);
			ptr->decCount();
		}
		
		//! Read this variable's content from config file and associate it to the GenericVarService name
		void associate(const std::string &name, T defaultValue)
		{
			assert(settings);
			
			// Find service
			Service *service = settings->findService(name);
			
			if (service == NULL)
			{
				// Service does not exists, create it
				ptr = new GenericVarService<T, ro, s>(name);
				settings->addService(ptr);
				
				// Set default value
				std::fill(&ptr->v[0], &ptr->v[s], defaultValue);
				
				// Get initial value from settings
				ParserTable::iterator i = settings->parsedTable.find(name);
				if (i != settings->parsedTable.end())
				{
					if (s == i->second.size())
					{
						for (Size j=0; j<i->second.size(); j++)
						{
							std::istringstream iss(i->second[j]);
							T tv;
							iss >> tv;
							ptr->v[j] = tv;
						}
					}
					else
					{
						std::cerr << "!!! Config entry " << name << " has wrong number of parameter" << std::endl;
					}
				}
				else if (warningWhenInitializedWithDefaultValue)
				{
					std::cout << "Ishtar::Variable::associate : service named " << name << " was initialized with default value (" << defaultValue << ")." << std::endl;
				}
			}
			else
			{
				ptr = dynamic_cast<GenericVarService<T, ro, s> *> (service);
				if (ptr)
				{
					ptr->incCount();
				}
				else
				{
					std::cerr << "!!! Ishtar::Variable::associate : a service named " << name << " of another type already exists !" << std::endl;
					assert(false);
				}
			}
		}
		
		//! Copy constructor
		Variable(const Variable<T, ro, s>& v) { ptr = v.ptr; ptr->incCount(); }
		
		//! Assignation operator
		Variable& operator=(const Variable<T, ro, s>& v) { GenericVarService<T, ro, s> *oPtr = ptr; ptr = v.ptr; ptr->incCount(); oPtr->decCount(); return *this; }
		
		//! Cast operator, return reference to v[0]
		operator T&() const
		{
			assert(ptr);
			return ptr->v[0];
		}
		
		//! Array access operator, return reference to v[index]
		T& operator[](Size index) const
		{
			assert(ptr);
			assert(index < s);
			return ptr->v[index];
		}
		
		//! Assignation operator, assign val to v[0]
		Variable & operator=(const T &val)
		{
			assert(ptr);
			ptr->v[0] = val;
			return *this;
		}
		
		//! Assignation operator, assign array val to v
		Variable & operator=(const T val[s])
		{
			assert(ptr);
			memcpy(ptr->v, val, s * sizeof(T));
			return *this;
		}
	};
	
	//! A variable for a string that behaves like a standard C++ variable excepted that its content is stored in a GenericVarService pointed by ptr and is accessible through network
	template<bool ro, bool warningWhenInitializedWithDefaultValue>
	class Variable<std::string, ro, 1, warningWhenInitializedWithDefaultValue>
	{
	private:
		//! The service really containing the value
		GenericVarService<std::string, ro, 1> *ptr;
		
	public:
		//! Constructor, associate this variable with name
		Variable(const std::string &name, const std::string &defaultValue = "")
		{
			associate(name, defaultValue);
		}
		
		//! Destructor
		~Variable()
		{
			assert(ptr);
			ptr->decCount();
		}
		
		//! Read this variable's content from config file and associate it to the GenericVarService name
		void associate(const std::string &name, const std::string &defaultValue)
		{
			// Find service
			Service *service = settings->findService(name);
			
			if (service == NULL)
			{
				// Service does not exists, create it
				ptr = new GenericVarService<std::string, ro, 1>(name);
				settings->addService(ptr);
				
				// Set default value
				ptr->v = defaultValue;
				
				// Get initial value from settings
				ParserTable::iterator i = settings->parsedTable.find(name);
				if (i != settings->parsedTable.end())
				{
					if (i->second.size() == 1)
					{
						ptr->v = i->second[0];
					}
					else
					{
						std::cerr << "!!! Config entry " << name << " should be of type string and thus have only one parameter" << std::endl;
					}
				}
				else if (warningWhenInitializedWithDefaultValue)
				{
					std::cout << "Ishtar::Variable::associate : service named " << name << " was initialized with default value (" << defaultValue << ")." << std::endl;
				}
			}
			else
			{
				ptr = dynamic_cast<GenericVarService<std::string, ro, 1> *> (service);
				if (ptr)
				{
					ptr->incCount();
				}
				else
				{
					std::cerr << "!!! Ishtar::Variable::associate : a service named " << name << " of another type already exists !" << std::endl;
					assert(false);
				}
			}
		}
		
		//! Copy constructor
		Variable(const Variable<std::string, ro, 1>& v) { ptr = v.ptr; ptr->incCount(); }
		
		//! Assignation operator
		Variable& operator=(const Variable<std::string, ro, 1>& v) { GenericVarService<std::string, ro, 1> *oPtr = ptr; ptr = v.ptr; ptr->incCount(); oPtr->decCount(); return *this; }
		
		//! Cast operator, return reference to v
		operator std::string &() const
		{
			assert(ptr);
			return ptr->v;
		}
		
		//! Cast operator, return v
		operator const char *() const
		{
			assert(ptr);
			return ptr->v.c_str();
		}
		
		//! Assignation operator, assign s to v
		Variable & operator=(const std::string &s)
		{
			assert(ptr);
			ptr->v = s;
			return *this;
		}
		
		//! Assignation operator, assign s to v
		Variable & operator=(const char *s)
		{
			assert(ptr);
			ptr->v = s;
			return *this;
		}
	};
	
	//! A vector that can accumulate values through + operator
	template<typename T>
	class ValuesAccumulator : public std::vector<T>
	{
	public:
		//! Add a new string to vector
		ValuesAccumulator &operator +(const T &v) { push_back(v); return *this; }
	};
	
	//! A vector of strings, used to pass subnames to FastVarService
	typedef ValuesAccumulator<std::string> StringsAccumulator;
	
	//! Expose some variable to Ishtar in a "light" way that doesn't change their behaviour outside Ishtar use
	template<typename T, bool ro = false, Size s = 1, bool warningWhenInitializedWithDefaultValue = false>
	class VarAccessService : public Service
	{
	protected:
		//! pointer to real values that I expose to Ishtar
		T *valuesPtr;
		
	public:
		//! Constructor
		VarAccessService()
		{
			valuesPtr = NULL;
		}
		
		//! Association, register itself to the services registry
		void associate(const std::string &name, T *variables, const StringsAccumulator &subnames = StringsAccumulator())
		{
			description.name = name;
			description.type = TypeResolver<T>::type;
			description.flags = ro ? ServiceDescription::READ_ONLY : 0;
			description.length = s;
			
			// if each variable in the array is named
			if (subnames.size() != 0)
			{
				description.flags |= ServiceDescription::NAMED_VALUES;
				description.subnames.resize(subnames.size());
				for (size_t i = 0; i < subnames.size(); i++)
					description.subnames[i] = subnames[i];
			}
			
			valuesPtr = variables;
			
			// Get initial value from settings
			ParserTable::iterator i = settings->parsedTable.find(name);
			if (i != settings->parsedTable.end())
			{
				if (s == i->second.size())
				{
					for (Size j=0; j<i->second.size(); j++)
					{
						std::istringstream iss(i->second[j]);
						T tv;
						iss >> tv;
						valuesPtr[j] = tv;
					}
				}
				else
				{
					std::cerr << "!!! Config entry " << name << " has wrong number of parameter" << std::endl;
				}
			}
			else if (warningWhenInitializedWithDefaultValue)
			{
				std::cout << "Ishtar::VarAccessService::associate : service named " << name << " was not initialized with config file." << std::endl;
			}
			
			settings->addService(this);
		}
		
		//! Destructor, unregister itself from the services registry
		virtual ~VarAccessService()
		{
		}
		
		//! Read values from the pointed variable
		virtual ValueVector *getValues(void)
		{
			ValueVectorTemplate<T> *vvt = new ValueVectorTemplate<T>(s);
			std::copy(valuesPtr, valuesPtr+s, &vvt->value[0]);
			return vvt;
		}
		
		//! Write values to the pointed variable
		virtual void setValues(ValueVector *values)
		{
			ValueVectorTemplate<T> *vvt = dynamic_cast<ValueVectorTemplate<T> *>(values);
			assert(vvt->value.size() == s);
			if (vvt)
			{
				std::copy(&vvt->value[0], &vvt->value[s], valuesPtr);
			}
			else
			{
				const std::string &name = description.name;
				std::cerr << "!!! Wrong type passed to " << name << "::setValues" << std::endl;
			}
		}
	};
	
	//! Expose a list of variables to Ishtar using VarAccessService
	class VariablesPublisher
	{
	protected:
		//! the services for exposing the variables
		std::vector<Service *> accessors;
		//! The ishtar root of this publisher
		std::string root;
		
	public:
		//! Constructor
		VariablesPublisher(const std::string &root = "")
		{
			this->root = root;
		}
		
		//! Destructor
		virtual ~VariablesPublisher()
		{
			for (size_t i=0; i<accessors.size(); i++)
				settings->eraseService(accessors[i]);
		}
		
		//! Set the root of this publisher
		void setRoot(const std::string &root) { this->root = root; }
		
		//! Expose a new variable
		template<typename T, bool ro, Size sz, bool warningWhenInitializedWithDefaultValue>
		void add(T *var, const std::string &name, const StringsAccumulator &subnames = StringsAccumulator())
		{
			VarAccessService<T, ro, sz, warningWhenInitializedWithDefaultValue> *s = new VarAccessService<T, ro, sz, warningWhenInitializedWithDefaultValue>();
			s->associate(root + name, var, subnames);
			accessors.push_back(s);
		}
		
		//! Expose a new variable
		template<typename T, bool ro, Size sz>
		void add(T *var, const std::string &name, const StringsAccumulator &subnames = StringsAccumulator())
		{
			add<T, ro, sz, false>(var, name, subnames);
		}
		
		//! Expose a new variable
		template<typename T, bool ro>
		void add(T *var, const std::string &name, const StringsAccumulator &subnames = StringsAccumulator())
		{
			add<T, ro, 1>(var, name, subnames);
		}
		
		//! Expose a new variable
		template<typename T>
		void add(T *var, const std::string &name, const StringsAccumulator &subnames = StringsAccumulator())
		{
			add<T, false>(var, name, subnames);
		}
	};
	
	/*template<typename T, bool ro = false, Size s = 1>
	class FastVariable
	{
	protected:
		T var;
		FastVarService<T, ro, s> service;
		
	public:
		FastVariable(const std::string &name , const std::string &defaultValue = "")
		{
			service.associate(name, &var, NULL);
		}
	
		operator T&() { return var; }
		operator const T&() const { return var; }
	};*/
	
	//! Used by user program to get the root of Ishtar arrays
	class ArrayHelper
	{
	protected:
		//! the Ishtar root for this array
		std::string root;
		//! The maximum number of elements in this array. The array is not dense, so maxIndex is not the actual number of element
		Variable<unsigned, true, 1> maxIndex;
		
	public:
		//! Constructor
		ArrayHelper(const std::string &root)
			:maxIndex(root + ".count", 0)
		{
			this->root = root;
		}
		
		//! Get a new unique root
		std::string getNewRoot(void)
		{
			std::ostringstream oss;
			oss << root << "." << maxIndex;
			maxIndex = maxIndex + 1;
			// TODO : duplicate values from default value
			return oss.str();
		}
		
		//! Return the number the last element. It can be bigger that the maximum number of element because of fragmentation
		unsigned getMaxIndex(void) { return maxIndex; }
		
		//! Return the root for a given element. This root is virtual, it may be that no variable are associated with it
		std::string getRoot(unsigned index)
		{
			std::ostringstream oss;
			oss << root << "." << index;
			return oss.str();
		}
		
		//! Lurk in settings to get the maximum number of elements for this root
		size_t getSettingsCount(void)
		{
			size_t i = 0;
			while (true)
			{
				std::ostringstream oss;
				oss << root << "." << i;
				const std::string &rootToTest = oss.str();
				bool found = false;
				for (ParserTable::iterator it = settings->parsedTable.begin(); it != settings->parsedTable.end(); ++it)
				{
					if (it->first.compare(0, rootToTest.length(), rootToTest) == 0)
						found = true;
				}
				if (!found)
					break;
				i++;
			}
			return i;
		}
	};
}

#endif
